#include <iostream>
#include <fstream>
#include <cmath>
#include "config.h"

using namespace std;

typedef float Real;

const Real dt = 0.01;

bool do_output = false;

Real B[3] = { 0, 0, 1 };
Real E[3] = { 0, 0, 0 };

Real R[NPART][2];
Real U[NPART][3];

typedef int Particle;

const Real PI = 3.1415926535897932384626;
template<typename T> T sqr(T x){ return x*x; }

void advance_particle_kernel(Particle p) {
    Real El[3] = { E[0], E[1], E[2] };
    Real Bl[3] = { B[0], B[1], B[2] };

    Real qm = 1;
    Real qmpidt = qm * PI * dt;

    /// 1/gamma = 1 / sqrt( 1 + |U|^2 )
    Real igamma = 1 / sqrt( 1 + sqr(U[p][0])+sqr(U[p][1])+sqr(U[p][2]) );

    Real S[3] = { igamma * qmpidt * Bl[0],
                  igamma * qmpidt * Bl[1],
                  igamma * qmpidt * Bl[2] };
    Real Edif[3] = { qmpidt * El[0],
                     qmpidt * El[1],
                     qmpidt * El[2] };

    /// um = U+Edif
    Real um[3] = { U[p][0] + Edif[0],
                   U[p][1] + Edif[1],
                   U[p][2] + Edif[2] };

    /// u0 = um + cross(um,S)
    Real u0[3] = { um[0] + um[1]*S[2] - um[2]*S[1],
                   um[1] + um[2]*S[0] - um[0]*S[2],
                   um[2] + um[0]*S[1] - um[1]*S[0] };

    /// up = um + 2/(1+|S|^2) * cross(u0,S)
    Real Sc2 = 2 / ( 1 + sqr(S[0])+sqr(S[1])+sqr(S[2]) );
    Real up[3] = { um[0] + Sc2 * (u0[1]*S[2] - u0[2]*S[1]),
                   um[1] + Sc2 * (u0[2]*S[0] - u0[0]*S[2]),
                   um[2] + Sc2 * (u0[0]*S[1] - u0[1]*S[0]) };

    /// u = up + Edif
    Real u[3] = { up[0] + Edif[0],
                  up[1] + Edif[1],
                  up[2] + Edif[2] };



    U[p][0] = u[0];
    U[p][1] = u[1];
    U[p][2] = u[2];

    igamma = 1 / sqrt( 1 + sqr(u[0])+sqr(u[1])+sqr(u[2]) );

    /// R = R + dt*u
    R[p][0] += igamma * dt * u[0];
    R[p][1] += igamma * dt * u[1];

}

ofstream out_part, out_u;
void out_open(){
    char filename[255];
    sprintf(filename,"out-particles");
    out_part.open(filename);
    sprintf(filename,"out-u");
    out_u.open(filename);
}
void out_close(){
    out_part.close();
    out_u.close();
}
void out(int n){
    out_part << n;
    out_u << n;
    for(int p=0; p<NPART; ++p){
        out_part << " " << R[p][0] << " " << R[p][1];
        out_u    << " " << U[p][0] << " " << U[p][1] << " " << U[p][2];
    }
    out_part << "\n";
    out_u << "\n";
}

void initialize_particles(){
    int cols = ceil(sqrt(NPART));
    int col = 1; int row = 1;
    for (int p=0; p<NPART; ++p){

        U[p][0] = 0.1+0.01*p;
        U[p][1] = 0;
        U[p][2] = 0;

        R[p][0] = 0.2*col;
        R[p][1] = 0.2*row;

        ++col; if(col>cols){ col = 1; ++row; }
    }
}

int main(void) {
    out_open();
    initialize_particles();
    for (int n=0; n<NSTEP; ++n){
        for (int p=0; p<NPART; ++p){
            advance_particle_kernel(p);
        }
        if (do_output) out(n);
    }
    out_close();
    return 0;
}
