#include <iostream>
#include <fstream>
#include <cmath>
#include <array>
#include "config.h"
#include "expressions.h"

/************************* CURRENT STATE:
 *
 * addition, subtraction, scalar multiplication works
 * same goes for cross product, and probably norm2, and missing parentheses (another 4-6%)
 * cross product should be implemented
 * transparent support for various register lengths
 * SIMD should be further abstracted (allow operations like norm)
 * expression templates need all the operators, parentheses, etc.
 *
 */

using namespace std;

typedef std::array<SIMD, 2> V2;
typedef math_vector<3> V3;
typedef std::array<Real,3> OneVec;

const Real dt = 0.01;


bool do_output = false;

OneVec B = { 0, 0, 1 };
OneVec E = { 0.01, 0.01, 0 };

std::array< V2, NPART/REGSIZE > R;
std::array< V3, NPART/REGSIZE > U;

typedef int Particle;

const Real PI = 3.1415926535897932384626;
template<typename T> T sqr(T x){ return x*x; }

//template<typename T> T mm_add(T a, T b){ return  _mm_add_ps(a, b); }
//template<typename T> T mm_sub(T a, T b){ return  _mm_sub_ps(a, b); }
//template<typename T> T mm_mul(T a, T b){ return  _mm_mul_ps(a, b); }
//template<typename T> T mm_div(T a, T b){ return  _mm_div_ps(a, b); }
//template<typename T> T mm_sqrt(T a){     return _mm_sqrt_ps(a); }

inline static SIMD sqrt(SIMD a){ return _mm_sqrt_ps(a); }
template<typename R> SIMD mm_set1(R a){ return _mm_set1_ps(a); }

//inline VEC operator+(const VEC a, const VEC b) { return __mm_add_ps(a, b); }


//TODO: change this to accommodate different register sizes (i.e. double)
template <typename T>
inline static SIMD norm2(const T& a) {
    SIMD r =  { sqr(a[0][0]) + sqr(a[1][0])  + sqr(a[2][0]),
                sqr(a[0][1]) + sqr(a[1][1])  + sqr(a[2][1]),
                sqr(a[0][2]) + sqr(a[1][2])  + sqr(a[2][2]),
                sqr(a[0][3]) + sqr(a[1][3])  + sqr(a[2][3]),
        };
    return r;
}

//TODO: Create a template expression for this one
inline static V3 cross(const V3& a, const V3 &b){
    V3 cr = { a[1]*b[2] - a[2]*b[1],
              a[2]*b[0] - a[0]*b[2],
              a[0]*b[1] - a[1]*b[0]
    };
    return cr;
}


const SIMD ones = { 1, 1, 1, 1 };
const SIMD twos = { 2, 2, 2, 2 };
const SIMD mPI  = { PI, PI, PI, PI };

void advance_particle_kernel(Particle p) {
    const SIMD mdt = mm_set1(dt);

    /// These shouldn't be "set1", but "set" with interpolated values at particles positions
    SIMD Ex = mm_set1(E[0]);
    SIMD Ey = mm_set1(E[1]);
    SIMD Ez = mm_set1(E[2]);

    SIMD Bx = mm_set1(B[0]);
    SIMD By = mm_set1(B[1]);
    SIMD Bz = mm_set1(B[2]);

    V3 El = { Ex, Ey, Ez };
    V3 Bl = { Bx, By, Bz };

    SIMD qm = mm_set1(1);

    SIMD qmpidt = qm * mPI * mdt;

    /// 1/gamma = 1 / sqrt( 1 + |U|^2 )
    SIMD igamma = ones / sqrt( ones + norm2(U[p]) );

    V3 Edif = qmpidt * El;
    V3 S = igamma * qmpidt * Bl;

    //std::cout << Edif[1][1] << std::endl;

    /// um = U+Edif
    V3 um = U[p]+Edif;

    /// u0 = um + cross(um,S)
    V3 u0 = um + cross(um,S);

    /// Sc2 = 2/(1+|S|^2)
    SIMD Sc2 = twos / ( ones + norm2(S) );
    /// up = um + Sc2 * cross(u0,S)
    V3 crs = cross(u0, S);
    V3 tms = Sc2 * crs;
    V3 up  = um + tms;
    //V3 up = um + (Sc2 * cross(u0, S));

    /// u = up + Edif
    V3 u = up + Edif;

    U[p] = u;

    SIMD newigamma = ones / sqrt(ones+norm2(u));

    /// R = R + dt*u
    R[p][0] += newigamma * mdt * u[0];
    R[p][1] += newigamma * mdt * u[1];
}

ofstream out_part, out_u;
void out_open(){
    char filename[255];
    sprintf(filename,"out-particles");
    out_part.open(filename);
    sprintf(filename,"out-u");
    out_u.open(filename);
}
void out_close(){
    out_part.close();
    out_u.close();
}
void out(int n){
    out_part << n;
    out_u << n;
    for(int p=0; p<NPART/REGSIZE; ++p){
        for (int r=0; r<REGSIZE; ++r){
            out_part << " " << R[p][0][r] << " " << R[p][1][r];
            out_u    << " " << U[p][0][r] << " " << U[p][1][r] << " " << U[p][2][r];
        }
    }
    out_part << "\n";
    out_u << "\n";
}

void initialize_particles(){
    int cols = ceil(sqrt(NPART));
    int col = 1; int row = 1;
    for(int p=0; p<NPART/REGSIZE; ++p){

        U[p][0][0] = 0.1+0.01*(p*REGSIZE);
        U[p][0][1] = 0.1+0.01*(p*REGSIZE+1);
        U[p][0][2] = 0.1+0.01*(p*REGSIZE+2);
        U[p][0][3] = 0.1+0.01*(p*REGSIZE+3);

        U[p][1][0] = 0;
        U[p][1][1] = 0;
        U[p][1][2] = 0;
        U[p][1][3] = 0;

        U[p][2][0] = 0;
        U[p][2][1] = 0;
        U[p][2][2] = 0;
        U[p][2][3] = 0;

        R[p][0][0] = 0.2*col;
        R[p][1][0] = 0.2*row;

        ++col; if(col>cols){ col = 1; ++row; }

        R[p][0][1] = 0.2*col;
        R[p][1][1] = 0.2*row;

        ++col; if(col>cols){ col = 1; ++row; }

        R[p][0][2] = 0.2*col;
        R[p][1][2] = 0.2*row;

        ++col; if(col>cols){ col = 1; ++row; }

        R[p][0][3] = 0.2*col;
        R[p][1][3] = 0.2*row;

        ++col; if(col>cols){ col = 1; ++row; }

    }
}

int main(void) {
    out_open();
    initialize_particles();
    for (int n=0; n<NSTEP; ++n){
        for (int p=0; p<NPART/REGSIZE; ++p){
            advance_particle_kernel(p);
        }
        if (do_output) out(n);
    }
    out_close();
    return 0;
}
